# Objet du script :
# Mise en œuvre d'un algorithme de Madgwick pour la fusion de données
# dans le cas d'une centrale à inertie mesurant 6 degrés de libertés (6DOF) :
# 3 d'accélération et 3 de rotation.
# Le capteur choisi est le LSM6DSO de STMicroelectronics et les axes Ox, Oy, Oz sont orientés
# comme sur le shield X-NUCLEO IKS01A3. 
# Cet exemple est une application directe des excellentes ressources partagées par Peter Hinch
# à cette URL : https://github.com/micropython-IMU/micropython-fusion.
# Le tutoriel de Peter Hinch propose deux variantes pour la fusion de capteur, une synchrone
# et une asynchrone. Nous avons choisi cette deuxième approche pour illustrer l'usage 
# la bibliothèque uasyncio de MicroPython.
# NB : le cap "Heading" reste égal à zéro car un magnétomètre est nécessaire pour le déterminer
# (voir l'exemple avec 9 degrés de liberté).

from machine import I2C # Pilote du bus I2C
import lsm6dso # Pilote de la centrale à inertie
from time import sleep_ms # Pour temporiser

import uasyncio as asyncio # Pour la gestion asynchrone
import gc # Appel du "ramasse miettes"
from fusion_async import Fusion # Pour la fusion de données

# On utilise l'I2C n°1 de la carte NUCLEO-W55 pour communiquer avec le capteur
i2c = I2C(1)

# Pause d'une seconde pour laisser à l'I2C le temps de s'initialiser
sleep_ms(1000)

# Liste des adresses I2C des périphériques présents
print("Adresses I2C utilisées : " + str(i2c.scan()))

# Instanciation du capteur
intertial_sensor = lsm6dso.LSM6DSO(i2c)

# Moindre sensibilité pour les mesures angulaires
intertial_sensor.scale_g('2000')

# Sensibilité maximum pour les mesures d'accélérations
intertial_sensor.scale_a('2g') 

# Coroutine pour la collecte des données et le paramétrage de sa fréquence
async def read_coro():
	
	# Temporisation
	await asyncio.sleep_ms(20)

	# Mesures des accélérations selon les 3 axes 
	ax = intertial_sensor.ax()
	ay = -1 * intertial_sensor.ay()
	az = intertial_sensor.az()

	# Construction du tuple des accélérations
	accel = (ax,ay,az)

	# Mesures des angles de rotation selon les 3 axes (en degrés/s)
	# Il est impératif que ces valeurs soient exprimées en degrés/s
	# et pas en radians/s !
	gx = intertial_sensor.gx() / 1000
	gy = -1 * intertial_sensor.gy() / 1000
	gz = intertial_sensor.gz() / 1000
	
	# Construction du tuple des gyrations
	gyro = (gx, gy, gz)

	# Affichage sur le port série
#	print("Accélérations")
#	print("Ax (mg) = %d" %accel[0])
#	print("Ay (mg) = %d" %accel[1])
#	print("Az (mg) = %d" %accel[2])
#	print("")
	
#	print("Rotations")
#	print("Gx (°/s) = %d" %gyro[0])
#	print("Gy (°/s) = %d" %gyro[1])
#	print("Gz (°/s) = %d" %gyro[2])
#	print("")

	return accel, gyro

# Instanciation pour la fusion des données
fuse = Fusion(read_coro)

# Couroutine d'appels forcés et réguliers au ramasse-miettes 
# Nécessaire pour la stabilité sur le long terme
async def mem_manage():
	while True:
		await asyncio.sleep_ms(100)
		gc.collect()
		gc.threshold(gc.mem_free() // 4 + gc.mem_alloc())
		

# Couroutine pour l'affichage
async def display():
	fs = 'Heading: {:4.0f} Pitch: {:4.0f} Roll: {:4.0f}'
	while True:
		print(fs.format(fuse.heading, fuse.pitch, fuse.roll))
		await asyncio.sleep_ms(100)

# Couroutine pour la fusion des données
async def fuse_task():
	await fuse.start()
	loop = asyncio.get_event_loop()
	await display()

# Création et lancement des tâches 
asyncio.create_task(mem_manage())
asyncio.run(fuse_task())