# Objet du script :
# Mise en œuvre d'un algorithme de Madgwick pour la fusion de données
# dans le cas d'une centrale à inertie mesurant 9 degrés de libertés (9DOF) :
# 3 d'accélération, 3 de rotation et 3 de champ magnétique.
# Les capteurs choisis sont le LSM6DSO et le LIS2MDL de STMicroelectronics. 
# Leurs axes Ox, Oy, Oz respectifs sont orientés comme sur le shield X-NUCLEO IKS01A3. 
# Cet exemple est une application directe des excellentes ressources partagées par Peter Hinch
# à cette URL : https://github.com/micropython-IMU/micropython-fusion.
# Le tutoriel de Peter Hinch propose deux variantes pour la fusion de capteur, une synchrone
# et une asynchrone. Nous avons choisi cette deuxième approche pour illustrer l'usage 
# la bibliothèque uasyncio de MicroPython.
# On suppose que les constantes de calibrage du magnétomètre ont été préalablement déterminées par le script
# dédié à cette opération dans le tutoriel https://stm32python.gitlab.io/fr/docs/Micropython/IKS01A3/lis2mdl.

from machine import I2C # Pilote du bus I2C
import lsm6dso # Pilote de la centrale à inertie
import lis2mdl # Pilote du magnétomètre
from time import sleep_ms # Pour temporiser

import uasyncio as asyncio # Pour la gestion asynchrone
import gc # Appel du "ramasse miettes"
from fusion_async import Fusion # Pour la fusion de données

# On utilise l'I2C n°1 de la carte NUCLEO-W55 pour communiquer avec le capteur
i2c = I2C(1)

# Pause d'une seconde pour laisser à l'I2C le temps de s'initialiser
sleep_ms(1000)

# Liste des adresses I2C des périphériques présents
print("Adresses I2C utilisées : " + str(i2c.scan()))

# Instanciation du gyroscope et de l'accéléromètre
intertial_sensor = lsm6dso.LSM6DSO(i2c)

# Instanciation du magnétomètre
magnetometer = lis2mdl.LIS2MDL(i2c)

# Moindre sensibilité pour les mesures angulaires
intertial_sensor.scale_g('2000')

# Sensibilité maximum pour les mesures d'accélérations
intertial_sensor.scale_a('2g') 

# Valeurs extrêmes selon chaque axe du champ magnétique obtenues par le script de calibration
# de la page https://stm32python.gitlab.io/fr/docs/Micropython/IKS01A3/lis2mdl.

min_bx = const(-75)
max_bx = const(18)
min_by = const(-24)
max_by = const(61)
min_bz = const(-120)
max_bz = const(-42)

# Calcul des décalages (offsets) "Hard Iron" pour chaque axe :

offset_x = (max_bx + min_bx) / 2
offset_y = (max_by + min_by) / 2
offset_z = (max_bz + min_bz) / 2

# Calcul des coefficients pour la correction approximative des 
# distorsions "Soft Iron" pour chaque axe :

avg_delta_x = (max_bx - min_bx) / 2
avg_delta_y = (max_by - min_by) / 2
avg_delta_z = (max_bz - min_bz) / 2

avg_delta = (avg_delta_x + avg_delta_y + avg_delta_z) / 3

scale_x = avg_delta / avg_delta_x
scale_y = avg_delta / avg_delta_y
scale_z = avg_delta / avg_delta_z

# Coroutine pour la collecte des données et le paramétrage de la fréquence
async def read_coro():
	
	# Temporisation
	await asyncio.sleep_ms(20)

	# Mesures des accélérations selon les 3 axes 
	ax = intertial_sensor.ax()
	ay = -1 * intertial_sensor.ay()
	az = intertial_sensor.az()

	# Mesures des angles de rotation selon les 3 axes (en degrés/s)
	# Il est impératif que ces valeurs soient exprimées en degrés/s
	# et pas en radians/s !
	gx = intertial_sensor.gx() / 1000
	gy = -1 * intertial_sensor.gy() / 1000
	gz = intertial_sensor.gz() / 1000
	
	# Mesure du champ magnétique selon les 3 axes
	# Puis on applique les corrections hard iron et soft iron.
	bx = (magnetometer.x() - offset_x) * scale_x
	by = (magnetometer.y() - offset_y) * scale_y
	bz = (magnetometer.z() - offset_z) * scale_z
	
	# Renvoie les mesures sous forme de rumples
	return (ax,ay,az),(gx, gy, gz),(bx, by, bz)

# Instanciation pour la fusion des données
fuse = Fusion(read_coro)

# Couroutine d'appels forcés et réguliers au ramasse-miettes 
# Nécessaire pour la stabilité sur le long terme
async def mem_manage():
	while True:
		await asyncio.sleep_ms(100)
		gc.collect()
		gc.threshold(gc.mem_free() // 4 + gc.mem_alloc())

# Couroutine pour l'affichage
async def display():
	fs = 'Heading: {:4.0f} Pitch: {:4.0f} Roll: {:4.0f}'
	while True:

		# Si le cap obtenu est négatif, 
		# calcule son complément à 360° 
		if fuse.heading < 0:
			fuse.heading += 360

		print(fs.format(fuse.heading, fuse.pitch, fuse.roll))
		await asyncio.sleep_ms(100)

# Couroutine pour la fusion des données
async def fuse_task():
	await fuse.start()
	loop = asyncio.get_event_loop()
	await display()

# Création et lancement des tâches 
asyncio.create_task(mem_manage())
asyncio.run(fuse_task())